@@ -27,8 +27,10 @@
 .storage <- function(x)
   attr(x, ".env");
 
-.get <- function(x, name, default, allow.null = FALSE) {
+.get <- function(x, name, default,
+                 allow.null = FALSE) {
   ret <- NULL
+
   # Make sure default is evaluated only when necessary
   if (.exists(x, name))
     ret <- get(name, envir = .storage(x))
@@ -51,9 +53,6 @@
 .exists <- function(x, name)
   exists(name, envir = .storage(x), inherits = FALSE);
 
-.is.extptrnull <- function(x)
-  .Call("is_extptrnull", x)
-
 .exists.non.null <- function(x, name) {
   ret <- FALSE
 
@@ -67,6 +66,13 @@
 .remove <- function(x, name)
   rm(list = name, envir = .storage(x), inherits = FALSE);
 
+.deprecate <- function(x, name, instead = NULL) {
+  val <- NA
+  attr(val, "deprecated") <- TRUE
+  attr(val, "instead") <- instead
+  .set(x, name, val)
+}
+
 .clone <- function(x, copy.storage = TRUE) {
   # Copy the information body
   obj <- x;
@@ -85,7 +91,7 @@
     }
   }
 
-  .init(obj)
+  obj
 }
 
 .get.series.info <- function(x) {
@@ -117,6 +123,61 @@
   F
 }
 
+.F <- function(x)
+  .get(x, "F")
+
+.decomposition <- function(x, field) {
+  d <- .get(x, "decomposition", allow.null = TRUE)
+  if (missing(field)) d else if (length(field) == 1) d[[field]] else d[field]
+}
+
+.set.decomposition <- function(x, ..., kind = "ssa.decomposition") {
+  val <- list(...)
+  class(val) <- kind
+
+  d <- .get(x, "decomposition", allow.null = TRUE)
+  if (is.null(d))
+    d <- list()
+
+  .set(x, "decomposition", modifyList(d, val))
+}
+
+.U.default <- function(x)
+  x$U
+
+.colspan.default <- function(x, idx)
+  x$U[, idx, drop = FALSE]
+
+.V.default <- function(x)
+  x$V
+
+.rowspan.default <- function(x, idx)
+  x$V[, idx, drop = FALSE]
+
+.sigma.default <- function(x)
+  x$sigma
+
+.U.ssa <- function(x)
+  .decomposition(x, "U")
+
+.colspan.ssa <- function(x, idx)
+  .colspan(.decomposition(x), idx)
+
+.V.ssa <- function(x)
+  .decomposition(x, "V")
+
+.rowspan.ssa <- function(x, idx)
+  .rowspan(.decomposition(x), idx)
+
+.sigma.ssa <- function(x)
+  .decomposition(x, "sigma")
+
+nspecial.ssa <- function(x)
+  return(0)
+
+.is.extptrnull <- function(x)
+  .Call("is_extptrnull", x)
+
 .na.omit <- function(x, ...) {
   # Drop initial and final NAs
   good <- which(!is.na(x))
@@ -259,6 +320,18 @@ wnorm <- function(x, ...)
   UseMethod(".hankelize.one")
 .hankelize.multi <- function(x, ...)
   UseMethod(".hankelize.multi")
+.U <- function(x, ...)
+  UseMethod(".U")
+.colspan <- function(x, ...)
+  UseMethod(".colspan")
+.V <- function(x, ...)
+  UseMethod(".V")
+.rowspan <- function(x, ...)
+  UseMethod(".rowspan")
+.sigma <- function(x, ...)
+  UseMethod(".sigma")
+nspecial <- function(x)
+  UseMethod("nspecial")
 .elseries <- function(x, ...)
   UseMethod(".elseries")
 .init <- function(x, ...)
