@@ -19,15 +19,18 @@
 
 #   Routines for toeplitz SSA
 
-Lcor <- function(F, L) {
+Lcor <- function(F, L, circular = FALSE) {
   storage.mode(F) <- "double"
   storage.mode(L) <- "integer"
-  .Call("Lcor", F, L)
+  storage.mode(circular) <- "logical"
+  .Call("Lcor", F, L, circular)
 }
 
-new.tmat <- function(F, L = (N + 1) %/% 2, fft.plan = NULL) {
+new.tmat <- function(F, L = (N + 1) %/% 2,
+                     circular = FALSE,
+                     fft.plan = NULL) {
   N <- length(F)
-  R <- Lcor(F, L)
+  R <- Lcor(F, L, circular = circular)
 
   storage.mode(R) <- "double"
   t <- .Call("initialize_tmat", R, if (is.null(fft.plan)) fft.plan.1d(2*L - 1) else fft.plan)
@@ -51,17 +54,11 @@ tmatmul <- function(tmat, v, transposed = FALSE) {
   .Call("tmatmul", tmat, v, transposed)
 }
 
-.init.toeplitz.ssa <- function(x, ...) {
-  # Initialize FFT plan
-  .get.or.create.fft.plan(x)
-
-  x
-}
-
 .hankelize.one.toeplitz.ssa <- .hankelize.one.1d.ssa
 
 .get.or.create.tmat <- function(x) {
-  .get.or.create(x, "tmat", new.tmat(F = x$F, L = x$window))
+  .get.or.create(x, "tmat", new.tmat(F = x$F, L = x$window,
+                                     circular = x$circular))
 }
 
 decompose.toeplitz.ssa.nutrlan <- function(x,
@@ -69,34 +66,31 @@ decompose.toeplitz.ssa.nutrlan <- function(x,
                                            ...) {
   N <- x$length; L <- x$window; K <- N - L + 1
 
-  F <- .get(x, "F")
+  F <- .F(x)
   h <- .get.or.create.hmat(x)
 
-  olambda <- .get(x, "olambda", allow.null = TRUE)
-  U <- .get(x, "U", allow.null = TRUE)
+  lambda <- .decomposition(x)$lambda
+  U <- .U(x)
 
   T <- .get.or.create.tmat(x)
 
   S <- trlan.eigen(T, neig = neig, ...,
-                   lambda = olambda, U = U)
+                   lambda = lambda, U = U)
 
   # Save results
-  .set(x, "olambda", S$d)
-  if (!is.null(S$u))
-    .set(x, "U", S$u)
-
   num <- length(S$d)
-  lambda <- numeric(num)
+  sigma <- numeric(num)
   V <- matrix(nrow = K, ncol = num)
   for (i in 1:num) {
     Z <- hmatmul(h, S$u[, i], transposed = TRUE)
-    lambda[i] <- sqrt(sum(Z^2))
-    V[, i] <- Z / lambda[i]
+    sigma[i] <- sqrt(sum(Z^2))
+    V[, i] <- Z / sigma[i]
   }
 
   # Save results
-  .set(x, "lambda", lambda)
-  .set(x, "V", V)
+  .set.decomposition(x,
+                     sigma = sigma, U = S$u, V = V, lambda = S$d,
+                     kind = "toeplitz.decomposition")
 
   x
 }
@@ -108,30 +102,31 @@ decompose.toeplitz.ssa.eigen <- function(x,
   N <- x$length; L <- x$window; K <- N - L + 1
 
   # Check, whether continuation of decomposition is requested
-  if (!force.continue && nlambda(x) > 0)
+  if (!force.continue && nsigma(x) > 0)
     stop("Continuation of decompostion is not supported for this method.")
 
   # Build hankel matrix
-  F <- .get(x, "F")
+  F <- .F(x)
   h <- .get.or.create.hmat(x)
 
   # Do decomposition
-  C <- toeplitz(Lcor(F, L))
+  C <- toeplitz(Lcor(F, L, circular = x$circular))
   S <- eigen(C, symmetric = TRUE)
 
-  .set(x, "U", S$vectors[, 1:neig])
-
-  lambda <- numeric(L)
+  sigma <- numeric(L)
   V <- matrix(nrow = K, ncol = L)
   for (i in 1:L) {
     Z <- hmatmul(h, S$vectors[,i], transposed = TRUE)
-    lambda[i] <- sqrt(sum(Z^2))
-    V[, i] <- Z / lambda[i]
+    sigma[i] <- sqrt(sum(Z^2))
+    V[, i] <- Z / sigma[i]
   }
 
   # Save results
-  .set(x, "lambda", lambda[1:neig])
-  .set(x, "V", V[, 1:neig, drop = FALSE])
+  .set.decomposition(x,
+                     sigma = sigma[1:neig],
+                     U = S$vectors[, 1:neig, drop = FALSE],
+                     V = V[, 1:neig, drop = FALSE],
+                     kind = "toeplitz.decomposition")
 
   x
 }
@@ -143,30 +138,29 @@ decompose.toeplitz.ssa.svd <- function(x,
   N <- x$length; L <- x$window; K <- N - L + 1
 
   # Check, whether continuation of decomposition is requested
-  if (!force.continue && nlambda(x) > 0)
+  if (!force.continue && nsigma(x) > 0)
     stop("Continuation of decompostion is not supported for this method.")
 
   # Build hankel matrix
-  F <- .get(x, "F")
+  F <- .F(x)
   h <- .get.or.create.hmat(x)
 
   # Do decomposition
-  C <- toeplitz(Lcor(F, L))
+  C <- toeplitz(Lcor(F, L, circular = x$circular))
   S <- svd(C, nu = neig, nv = neig)
 
-  .set(x, "U", S$u)
-
-  lambda <- numeric(neig)
+  sigma <- numeric(neig)
   V <- matrix(nrow = K, ncol = neig)
   for (i in 1:neig) {
     Z <- hmatmul(h, S$u[,i], transposed = TRUE)
-    lambda[i] <- sqrt(sum(Z^2))
-    V[, i] <- Z / lambda[i]
+    sigma[i] <- sqrt(sum(Z^2))
+    V[, i] <- Z / sigma[i]
   }
 
   # Save results
-  .set(x, "lambda", lambda)
-  .set(x, "V", V)
+  .set.decomposition(x,
+                     sigma = sigma, U = S$u, V = V,
+                     kind = "toeplitz.decomposition")
 
   x
 }
@@ -178,36 +172,25 @@ decompose.toeplitz.ssa.propack <- function(x,
   N <- x$length; L <- x$window; K <- N - L + 1
 
   # Check, whether continuation of decomposition is requested
-  if (!force.continue && nlambda(x) > 0)
+  if (!force.continue && nsigma(x) > 0)
     stop("Continuation of decompostion is not yet implemented for this method.");
 
-  F <- .get(x, "F")
+  S <- propack.svd(.get.or.create.tmat(x), neig = neig, ...)
+
   h <- .get.or.create.hmat(x)
-
-  olambda <- .get(x, "olambda", allow.null = TRUE)
-  U <- .get(x, "U", allow.null = TRUE)
-
-  T <- .get.or.create.tmat(x)
-
-  S <- propack.svd(T, neig = neig, ...)
-
-  # Save results
-  .set(x, "olambda", S$d)
-  if (!is.null(S$u))
-    .set(x, "U", S$u)
-
   num <- length(S$d)
-  lambda <- numeric(num)
+  sigma <- numeric(num)
   V <- matrix(nrow = K, ncol = num)
   for (i in 1:num) {
     Z <- hmatmul(h, S$u[, i], transposed = TRUE)
-    lambda[i] <- sqrt(sum(Z^2))
-    V[, i] <- Z / lambda[i]
+    sigma[i] <- sqrt(sum(Z^2))
+    V[, i] <- Z / sigma[i]
   }
 
   # Save results
-  .set(x, "lambda", lambda)
-  .set(x, "V", V)
+  .set.decomposition(x,
+                     sigma = sigma, U = S$u, V = V, lambda = S$d,
+                     kind = "toeplitz.decomposition")
 
   x
 }
@@ -220,5 +203,8 @@ decompose.toeplitz.ssa <- function(x,
   stop("Unsupported SVD method for Toeplitz SSA!")
 }
 
-calc.v.toeplitz.ssa <- function(x, idx, ...)
-  x$V[, idx]
+calc.v.toeplitz.ssa <- calc.v.1d.ssa
+
+.rowspan.toeplitz.ssa <- function(x, idx) {
+  qr.Q(qr(.V(x)[, idx, drop = FALSE]))
+}
